@@ -98,6 +98,7 @@ impl Grid {
         if settings.indent.is_none()
             && settings.alignment_h.is_none()
             && settings.alignment_v.is_none()
+            && settings.span.is_none()
         {
             return;
         }
@@ -118,6 +119,9 @@ impl Grid {
         if let Some(alignment) = settings.alignment_v {
             s.alignment_v = alignment;
         }
+        if let Some(span) = settings.span {
+            s.span = span;
+        }
 
         self.styles.insert(entity, s);
     }
@@ -232,24 +236,8 @@ impl Grid {
         unreachable!("there's a global settings guaranted in the map")
     }
 
-    fn info(&self) -> Info<'_> {
-        let count_rows = self.count_rows();
-        let count_columns = self.count_columns();
-        let mut column_widths = vec![0; count_columns];
-        // default height is 1 as we consider empty string has height 1
-        //
-        // it's crusial since if the default height will be equal to 0
-        // cell line will be not present on the grid like this
-        //
-        //  default 0      default 1
-        //    +++            +++
-        //    +++            |||
-        //    +++            +++
-        //                   |||
-        //                   +++
-        let mut row_heights = vec![1; count_rows];
+    fn build_cells(&self, count_rows: usize, count_columns: usize) -> Vec<Vec<(Vec<&str>, Style)>> {
         let mut rows = Vec::with_capacity(count_rows);
-
         (0..count_rows).for_each(|row_index| {
             let mut row = Vec::with_capacity(count_columns);
 
@@ -258,25 +246,13 @@ impl Grid {
                 let cell: Vec<_> = content.lines().collect();
                 let style = self.style(row_index, column_index);
 
-                let content_height = cell.len();
-                let cell_height = content_height + style.indent.top + style.indent.bottom;
-
-                let content_width = string_width(content);
-                let cell_width = content_width + style.indent.left + style.indent.right;
-
-                column_widths[column_index] = max(column_widths[column_index], cell_width);
-                row_heights[row_index] = max(row_heights[row_index], cell_height);
                 row.push((cell, style));
             });
 
             rows.push(row);
         });
 
-        Info {
-            cells: rows,
-            row_heights,
-            column_widths,
-        }
+        rows
     }
 
     fn default_border() -> Border {
@@ -303,13 +279,6 @@ impl Grid {
     }
 }
 
-#[derive(Debug)]
-struct Info<'a> {
-    cells: Vec<Vec<(Vec<&'a str>, Style)>>,
-    row_heights: Vec<usize>,
-    column_widths: Vec<usize>,
-}
-
 /// Settings represent setting of a particular cell
 #[derive(Debug, Clone, Default)]
 pub struct Settings {
@@ -317,6 +286,7 @@ pub struct Settings {
     indent: Option<Indent>,
     alignment_h: Option<AlignmentHorizontal>,
     alignment_v: Option<AlignmentVertical>,
+    span: Option<usize>,
 }
 
 impl Settings {
@@ -353,6 +323,12 @@ impl Settings {
         self.alignment_v = Some(alignment);
         self
     }
+
+    /// Set the settings's span.
+    pub fn set_span(mut self, span: usize) -> Self {
+        self.span = Some(span);
+        self
+    }
 }
 
 /// Border structure represent all borders of a row
@@ -478,6 +454,7 @@ struct Style {
     indent: Indent,
     alignment_h: AlignmentHorizontal,
     alignment_v: AlignmentVertical,
+    span: usize,
 }
 
 impl Default for Style {
@@ -491,6 +468,7 @@ impl Default for Style {
                 right: 0,
                 top: 0,
             },
+            span: 1,
         }
     }
 }
@@ -563,32 +541,37 @@ impl AlignmentVertical {
 
 impl std::fmt::Display for Grid {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        let count_rows = self.count_rows();
+        let count_columns = self.count_columns();
+
         // It may happen when all cells removed via `remove_row`, `remove_column` methods
-        if self.count_rows() == 0 || self.count_columns() == 0 {
+        if count_rows == 0 || count_columns == 0 {
             return Ok(());
         }
 
-        let information = self.info();
+        let mut cells = self.build_cells(count_rows, count_columns);
+        let row_heights = rows_height(&cells, count_rows, count_columns);
+        let widths = __columns_width(&mut cells, count_rows, count_columns);
 
-        for (row_index, row) in information.cells.into_iter().enumerate() {
+        for (row_index, row) in cells.into_iter().enumerate() {
             let border = self
                 .border_styles
                 .get(row_index)
                 .expect("it's expected that grid has N styles where N is an amount of rows");
 
             if row_index == 0 {
-                build_line(f, &information.column_widths, &border.top_line)?;
+                build_split_line(f, &widths[row_index], &border.top_line)?;
             }
 
             build_row(
                 f,
                 row,
-                &information.column_widths,
-                information.row_heights[row_index],
+                &widths[row_index],
+                row_heights[row_index],
                 &border.inner,
             )?;
 
-            build_line(f, &information.column_widths, &border.bottom_line)?;
+            build_split_line(f, &widths[row_index], &border.bottom_line)?;
         }
 
         Ok(())
@@ -597,63 +580,59 @@ impl std::fmt::Display for Grid {
 
 fn build_row(
     f: &mut std::fmt::Formatter<'_>,
-    mut row: Vec<(Vec<&str>, Style)>,
+    row: Vec<(Vec<&str>, Style)>,
     widths: &[usize],
     height: usize,
     border: &LineStyle,
 ) -> fmt::Result {
-    let mut top_indents = Vec::with_capacity(row.len());
-    for (cell, style) in &row {
-        let content_height = cell.len();
-        let height = height - style.indent.top - style.indent.bottom;
+    for _line in 0..height {
+        build_line(f, row.len(), border, |f, column| {
+            let (cell, style) = &row[column];
+            let width = widths[column];
 
-        let indent = style.alignment_v.top_ident(height, content_height);
-        let indent = indent + style.indent.top;
-
-        top_indents.push(indent);
-    }
-
-    for line_index in 0..height {
-        write_option(f, border.left_intersection)?;
-
-        for (column_index, (cell, style)) in row.iter_mut().enumerate() {
-            if column_index != 0 {
-                write_option(f, border.intersection)?;
+            let top_indent = top_indent(cell, style, height);
+            if top_indent > _line {
+                return empty_line(f, width);
             }
 
-            let width = widths[column_index];
-
-            if top_indents[column_index] > line_index {
-                empty_line(f, width)?;
-                continue;
+            let cell_line_index = _line - top_indent;
+            let is_cell_has_this_line = cell.len() > cell_line_index;
+            if !is_cell_has_this_line {
+                return empty_line(f, width);
             }
 
-            let index = line_index - top_indents[column_index];
-            match cell.get(index) {
-                Some(s) => line(
-                    f,
-                    s,
-                    width,
-                    style.indent.left,
-                    style.indent.right,
-                    style.alignment_h,
-                )?,
-                None => {
-                    empty_line(f, width)?;
-                }
-            }
-        }
-
-        write_option(f, border.right_intersection)?;
-
-        writeln!(f)?;
+            let line_text = cell[cell_line_index];
+            line(
+                f,
+                line_text,
+                width,
+                style.indent.left,
+                style.indent.right,
+                style.alignment_h,
+            )
+        })?;
     }
 
     Ok(())
 }
 
+fn top_indent(cell: &[&str], style: &Style, height: usize) -> usize {
+    let height = height - style.indent.top;
+    let content_height = cell_height(cell, style) - style.indent.top - style.indent.bottom;
+    let indent = style.alignment_v.top_ident(height, content_height);
+    indent + style.indent.top
+}
+
 fn empty_line(f: &mut std::fmt::Formatter<'_>, n: usize) -> fmt::Result {
-    write!(f, "{: ^1$}", "", n)
+    write!(f, "{:1$}", "", n)
+}
+
+fn repeat_char(f: &mut std::fmt::Formatter<'_>, c: char, n: usize) -> fmt::Result {
+    if n > 0 {
+        write!(f, "{:1$}", c, n)
+    } else {
+        Ok(())
+    }
 }
 
 fn line(
@@ -664,29 +643,26 @@ fn line(
     right_indent: usize,
     alignment: AlignmentHorizontal,
 ) -> fmt::Result {
-    write!(f, "{}", " ".repeat(left_indent))?;
+    repeat_char(f, ' ', left_indent)?;
     alignment.align(f, text, width - left_indent - right_indent)?;
-    write!(f, "{}", " ".repeat(right_indent))?;
+    repeat_char(f, ' ', right_indent)?;
     Ok(())
 }
 
-fn build_line(
+fn build_line<F: Fn(&mut std::fmt::Formatter<'_>, usize) -> fmt::Result>(
     f: &mut std::fmt::Formatter<'_>,
-    cells_width: &[usize],
+    length: usize,
     border: &LineStyle,
+    writer: F,
 ) -> fmt::Result {
-    if border.is_empty() {
-        return Ok(());
-    }
-
     write_option(f, border.left_intersection)?;
 
-    for (i, w) in cells_width.iter().enumerate() {
-        write_option(f, border.main.map(|m| m.to_string().repeat(*w)))?;
-
-        if i != cells_width.len() - 1 {
+    for i in 0..length {
+        if i != 0 {
             write_option(f, border.intersection)?;
         }
+
+        writer(f, i)?;
     }
 
     write_option(f, border.right_intersection)?;
@@ -696,6 +672,20 @@ fn build_line(
     Ok(())
 }
 
+fn build_split_line(
+    f: &mut std::fmt::Formatter<'_>,
+    widths: &[usize],
+    border: &LineStyle,
+) -> fmt::Result {
+    if border.is_empty() {
+        return Ok(());
+    }
+
+    build_line(f, widths.len(), border, |f, i| {
+        write_option(f, border.main.map(|m| m.to_string().repeat(widths[i])))
+    })
+}
+
 fn write_option<D: Display>(f: &mut std::fmt::Formatter<'_>, text: Option<D>) -> fmt::Result {
     match text {
         Some(text) => write!(f, "{}", text),
@@ -716,10 +706,191 @@ fn string_width(text: &str) -> usize {
 }
 
 fn real_string_width(text: &str) -> usize {
-    text.lines()
-        .map(unicode_width::UnicodeWidthStr::width)
-        .max()
-        .unwrap_or(0)
+    text.lines().map(line_width).max().unwrap_or(0)
+}
+
+fn line_width(line: &str) -> usize {
+    unicode_width::UnicodeWidthStr::width(line)
+}
+
+fn __columns_width(
+    cells: &mut [Vec<(Vec<&str>, Style)>],
+    count_rows: usize,
+    count_columns: usize,
+) -> Vec<Vec<usize>> {
+    let mut widths = vec![vec! {0; count_columns}; count_rows];
+    (0..count_rows).for_each(|row| {
+        (0..count_columns).for_each(|column| {
+            let (cell, style) = &cells[row][column];
+            if is_cell_visible(&cells[row], column) {
+                widths[row][column] = cell_width(cell, style);
+            } else {
+                widths[row][column] = 0;
+            }
+        });
+    });
+
+    // check if we don't need to check all spans as it a heavy load function.
+    // it suppose to save us time and resources
+    if cells
+        .iter()
+        .any(|row| row.iter().any(|(_, style)| style.span > 1))
+    {
+        (1..count_columns + 1).for_each(|span| {
+            __adjust_width(&mut widths, cells, count_rows, count_columns, span);
+        });
+    } else {
+        __adjust_width(&mut widths, cells, count_rows, count_columns, 1);
+    }
+
+    // remove not visible cells to print everything correctly
+    (0..count_rows).for_each(|row| {
+        let mut n_removed = 0;
+        (0..count_columns).for_each(|column| {
+            if !is_cell_visible(&cells[row], column) {
+                widths[row].remove(column - n_removed);
+                cells[row].remove(column - n_removed);
+                n_removed += 1;
+            }
+        });
+    });
+
+    widths
+}
+
+fn __adjust_width(
+    widths: &mut [Vec<usize>],
+    cells: &[Vec<(Vec<&str>, Style)>],
+    count_rows: usize,
+    count_columns: usize,
+    span: usize,
+) {
+    (0..count_rows).for_each(|row| {
+        (0..count_columns)
+            .filter(|&column| cells[row][column].1.span == span)
+            .filter(|&column| is_cell_visible(&cells[row], column))
+            .for_each(|column| {
+                let (cell, style) = &cells[row][column];
+                let cell_width = cell_width(cell, style);
+                // calc other's width
+
+                let others_width = (0..count_rows)
+                    .filter(|&r| r != row)
+                    .filter(|&r| {
+                        cells[r][column..column + span]
+                            .iter()
+                            .map(|(_, style)| style.span)
+                            .sum::<usize>()
+                            <= span
+                    })
+                    .map(|r| row_width(&cells[r], &widths[r][column..column + span]))
+                    .max()
+                    .unwrap_or(0);
+
+                if cell_width > others_width {
+                    widths[row][column] = cell_width;
+
+                    (0..count_rows)
+                        .filter(|&r| r != row)
+                        .filter(|&r| {
+                            cells[r][column..column + span]
+                                .iter()
+                                .map(|(_, style)| style.span)
+                                .sum::<usize>()
+                                <= span
+                        }) // not sure if it's correct
+                        .for_each(|r| {
+                            inc_width_to_cells(
+                                &cells[r],
+                                &mut widths[r][column..column + span],
+                                cell_width,
+                            );
+                        });
+                } else {
+                    inc_width_to_cells(
+                        &cells[row],
+                        &mut widths[row][column..column + 1],
+                        others_width,
+                    );
+                }
+            });
+    });
+}
+
+fn is_cell_visible(row: &[(Vec<&str>, Style)], column: usize) -> bool {
+    !row[..column]
+        .iter()
+        .zip(column..)
+        .any(|((_, style), span)| style.span > span)
+}
+
+// relyes on fix_spans
+fn row_width(row: &[(Vec<&str>, Style)], widths: &[usize]) -> usize {
+    let w = (0..widths.len())
+        .filter(|&i| is_cell_visible(row, i))
+        .map(|i| widths[i])
+        .sum::<usize>();
+
+    if w == 0 {
+        return 0;
+    }
+
+    w + (0..widths.len())
+        .filter(|&i| is_cell_visible(row, i))
+        .count()
+        - 1
+}
+
+// relyes on fix_spans
+fn inc_width_to_cells(row: &[(Vec<&str>, Style)], widths: &mut [usize], width: usize) {
+    let a = row_width(row, &widths);
+    let diff = width - a;
+    (0..diff)
+        .zip(
+            (0..widths.len())
+                .filter(|&i| is_cell_visible(row, i))
+                .cycle(),
+        )
+        .collect::<Vec<_>>()
+        .iter()
+        .for_each(|&(_, i)| widths[i] += 1);
+}
+
+fn cell_width(cell: &[&str], style: &Style) -> usize {
+    let content_width = cell.iter().map(|l| line_width(l)).max().unwrap_or(0);
+    content_width + style.indent.left + style.indent.right
+}
+
+fn rows_height(
+    cells: &[Vec<(Vec<&str>, Style)>],
+    count_rows: usize,
+    count_columns: usize,
+) -> Vec<usize> {
+    // default height is 1 as we consider empty string has height 1
+    //
+    // it's crusial since if the default height will be equal to 0
+    // cell line will be not present on the grid like this
+    //
+    //  default 0      default 1
+    //    +++            +++
+    //    +++            |||
+    //    +++            +++
+    //                   |||
+    //                   +++
+    let mut row_heights = vec![1; count_rows];
+    (0..count_rows).for_each(|row_index| {
+        (0..count_columns).for_each(|column_index| {
+            let (cell, style) = &cells[row_index][column_index];
+            row_heights[row_index] = max(row_heights[row_index], cell_height(cell, style));
+        });
+    });
+
+    row_heights
+}
+
+fn cell_height(cell: &[&str], style: &Style) -> usize {
+    let content_height = cell.len();
+    content_height + style.indent.top + style.indent.bottom
 }
 
 #[cfg(test)]
@@ -1010,4 +1181,20 @@ mod tests {
         assert_eq!(AlignmentVertical::Center.top_ident(3, 1), 1);
         assert_eq!(AlignmentVertical::Center.top_ident(4, 1), 1);
     }
+
+    #[test]
+    fn grid_2x2_span_test() {
+        let mut grid = Grid::new(2, 2);
+        grid.set(Entity::Global, Settings::new().text("asd"));
+        grid.set(Entity::Cell(0, 0), Settings::new().text("123").set_span(2));
+        let str = grid.to_string();
+        assert_eq!(
+            str,
+            "+-------+\n\
+             |123    |\n\
+             +-------+\n\
+             |asd|asd|\n\
+             +---+---+\n"
+        )
+    }
 }
